#
# MIT License
#
# (C) Copyright [2024] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Public API module for the kvm cluster layer, this gives callers
access to the Cluster API and prevents them from seeing the private
implementation of the API.

"""
from abc import (
    ABCMeta,
    abstractmethod
)
from ..layerapi import LayerAPI


class ClusterAPI(LayerAPI, metaclass=ABCMeta):
    """ Presents the Cluster API to callers.

    """
    @abstractmethod
    def get_virtual_nodes(self):
        """Return a VirtualNodes object containing all of the
        available non-pure-base-class Virtual Nodes.

        """

    @abstractmethod
    def get_virtual_networks(self):
        """Return a VirtualNetworks object containing all the
        available VirtualNetworks.

        """
